#include <stdarg.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdlib.h>

enum CharacterCategory {
  Body,
  Hair,
  Face,
  Tail,
  Ear,
};
typedef uint8_t CharacterCategory;

/**
 * Where the existing installation came from
 */
typedef enum ExistingInstallType {
  /**
   * Installed via the official launcher
   */
  OfficialLauncher,
  /**
   * Installed via XIVQuickLauncher
   */
  XIVQuickLauncher,
  /**
   * Installed via XIVLauncherCore
   */
  XIVLauncherCore,
  /**
   * Installed via XIVOnMac
   */
  XIVOnMac,
  /**
   * Installed via Astra
   */
  Astra,
} ExistingInstallType;

/**
 * Gender of the character.
 */
enum Gender {
  Male,
  Female,
};
typedef uint8_t Gender;

/**
 * The language the game data is written for. Some of these languages are supported in the Global region.
 */
enum Language {
  /**
   * Used for data that is language-agnostic, such as item data.
   */
  None,
  /**
   * Japanese language.
   */
  Japanese,
  /**
   * English language.
   */
  English,
  /**
   * German language.
   */
  German,
  /**
   * French language.
   */
  French,
  /**
   * Chinese (Simplified) language.
   */
  ChineseSimplified,
  /**
   * Chinese (Traditional) language.
   */
  ChineseTraditional,
  /**
   * Korean language.
   */
  Korean,
};
typedef uint8_t Language;

/**
 * The major races of Eorzea.
 */
enum Race {
  Hyur,
  Elezen,
  Lalafell,
  Miqote,
  Roegadyn,
  AuRa,
  Hrothgar,
  Viera,
};
typedef uint8_t Race;

typedef enum ShaderStage {
  Vertex,
  Pixel,
} ShaderStage;

/**
 * The slot the item is for.
 */
enum Slot {
  /**
   * The head slot. Shorthand is "met".
   */
  Head,
  /**
   * The hands slot. Shorthand is "glv".
   */
  Hands,
  /**
   * The legs slot. Shorthand is "dwn".
   */
  Legs,
  /**
   * The feet slot. Shorthand is "sho".
   */
  Feet,
  /**
   * The body or chest slot. Shorthand is "top".
   */
  Body,
  /**
   * The earrings slot. Shorthand is "ear".
   */
  Earring,
  /**
   * The neck slot. Shorthand is "nek".
   */
  Neck,
  /**
   * The ring slot. Shorthand is "rir".
   */
  Rings,
  /**
   * The wrists slot. Shorthand is "wrs".
   */
  Wrists,
};
typedef uint8_t Slot;

/**
 * The race's "subrace". Each race has two subraces, which are actually identical (even down to the ids!)
 * with the exception of Hyurs, which have two unique subraces that are really two separate races.
 */
enum Subrace {
  Midlander,
  Highlander,
  Wildwood,
  Duskwight,
  Plainsfolk,
  Dunesfolk,
  Seeker,
  Keeper,
  SeaWolf,
  Hellsguard,
  Raen,
  Xaela,
  Hellion,
  Lost,
  Rava,
  Veena,
};
typedef uint8_t Subrace;

typedef enum TextureType {
  TwoDimensional,
  ThreeDimensional,
} TextureType;

enum TextureUsage {
  Sampler,
  Sampler0,
  Sampler1,
  SamplerCatchlight,
  SamplerColorMap0,
  SamplerColorMap1,
  SamplerDiffuse,
  SamplerEnvMap,
  SamplerMask,
  SamplerNormal,
  SamplerNormalMap0,
  SamplerNormalMap1,
  SamplerReflection,
  SamplerSpecular,
  SamplerSpecularMap0,
  SamplerSpecularMap1,
  SamplerWaveMap,
  SamplerWaveletMap0,
  SamplerWaveletMap1,
  SamplerWhitecapMap,
  UnknownDawntrail1,
};
typedef uint8_t TextureUsage;

/**
 * Implementation of the Blowfish block cipher, specialized for encrypting and decrypting SqexArg - the technique used to encrypt game arguments by the launcher.
 *
 * # Example
 *
 * ```
 * # use physis::blowfish::Blowfish;
 * let key = b"abcdefgh";
 * let data = b"foobar  ";
 *
 * let fish = Blowfish::new(key);
 * let encrypted = fish.encrypt(data).unwrap();
 * let decrypted = fish.decrypt(&encrypted).unwrap();
 * # assert_eq!(data, &decrypted[..])
 * ```
 */
typedef struct Blowfish Blowfish;

/**
 * Represents the boot data for FFXIV, which is located under the "boot" directory.
 */
typedef struct BootData BootData;

typedef struct CMP CMP;

/**
 * Represents a config file, which is made up of categories and settings. Categories may have zero to one setting.
 */
typedef struct ConfigFile ConfigFile;

typedef struct EXD EXD;

typedef struct EXH EXH;

/**
 * Framework for operating on game data.
 */
typedef struct GameData GameData;

typedef struct MDL MDL;

typedef struct PreBoneDeformer PreBoneDeformer;

typedef struct Schd Schd;

typedef struct ShaderPackage ShaderPackage;

typedef struct physis_Buffer {
  uint32_t size;
  uint8_t *data;
} physis_Buffer;

/**
 * The type of repository, discerning game data from expansion data.
 */
typedef enum RepositoryType_Tag {
  /**
   * The base game directory, like "ffxiv".
   */
  Base,
  /**
   * An expansion directory, like "ex1".
   */
  Expansion,
} RepositoryType_Tag;

typedef struct Expansion_Body {
  /**
   * The expansion number starting at 1.
   */
  int32_t number;
} Expansion_Body;

typedef struct RepositoryType {
  RepositoryType_Tag tag;
  union {
    Expansion_Body expansion;
  };
} RepositoryType;

typedef struct physis_Repository {
  const char *name;
  struct RepositoryType repository_type;
  const char *version;
} physis_Repository;

typedef struct physis_Repositories {
  uint32_t repositories_count;
  struct physis_Repository *repositories;
} physis_Repositories;

typedef struct physis_EXH {
  struct EXH *p_ptr;
  uint32_t page_count;
  uint32_t language_count;
  Language *languages;
  uint32_t column_count;
  uint32_t row_count;
} physis_EXH;

typedef enum physis_ColumnData_Tag {
  String,
  Bool,
  Int8,
  UInt8,
  Int16,
  UInt16,
  Int32,
  UInt32,
  Float32,
  Int64,
  UInt64,
} physis_ColumnData_Tag;

typedef struct physis_ColumnData {
  physis_ColumnData_Tag tag;
  union {
    struct {
      const char *string;
    };
    struct {
      bool bool_;
    };
    struct {
      int8_t int8;
    };
    struct {
      uint8_t u_int8;
    };
    struct {
      int16_t int16;
    };
    struct {
      uint16_t u_int16;
    };
    struct {
      int32_t int32;
    };
    struct {
      uint32_t u_int32;
    };
    struct {
      float float32;
    };
    struct {
      int64_t int64;
    };
    struct {
      uint64_t u_int64;
    };
  };
} physis_ColumnData;

typedef struct physis_ExcelRow {
  struct physis_ColumnData *column_data;
} physis_ExcelRow;

typedef struct physis_EXD {
  struct EXD *p_ptr;
  unsigned int column_count;
  struct physis_ExcelRow *row_data;
  unsigned int row_count;
} physis_EXD;

typedef struct physis_SheetNames {
  uint32_t name_count;
  const char **names;
} physis_SheetNames;

typedef struct Vertex {
  float position[3];
  float uv0[2];
  float uv1[2];
  float normal[3];
  float bitangent[4];
  float color[4];
  float bone_weight[4];
  uint8_t bone_id[4];
} Vertex;

typedef struct SubMesh {
  uintptr_t submesh_index;
  uint32_t index_count;
  uint32_t index_offset;
} SubMesh;

typedef struct physis_Shape {
  const char *name;
  const struct Vertex *morphed_vertices;
} physis_Shape;

typedef struct physis_Part {
  uint32_t num_vertices;
  const struct Vertex *vertices;
  uint32_t num_indices;
  const uint16_t *indices;
  uint16_t material_index;
  uint32_t num_submeshes;
  const struct SubMesh *submeshes;
  uint32_t num_shapes;
  struct physis_Shape *shapes;
} physis_Part;

typedef struct physis_LOD {
  uint32_t num_parts;
  const struct physis_Part *parts;
} physis_LOD;

typedef struct physis_MDL {
  struct MDL *p_ptr;
  uint32_t num_lod;
  const struct physis_LOD *lods;
  uint32_t num_affected_bones;
  const char **affected_bone_names;
  uint32_t num_material_names;
  const char **material_names;
} physis_MDL;

typedef struct NewShapeValue {
  uint32_t base_index;
  struct Vertex replacing_vertex;
} NewShapeValue;

typedef struct physis_SupportedSubraces {
  Subrace subraces[2];
} physis_SupportedSubraces;

typedef struct physis_EXL {
  int32_t version;
  int32_t entry_count;
  const char *const *entry_keys;
  const int32_t *entry_values;
} physis_EXL;

typedef struct physis_Bone {
  uint32_t index;
  const char *name;
  struct physis_Bone *parent_bone;
  uint32_t parent_index;
  float position[3];
  float rotation[4];
  float scale[3];
} physis_Bone;

typedef struct physis_Skeleton {
  uint32_t num_bones;
  struct physis_Bone *bones;
  struct physis_Bone *root_bone;
} physis_Skeleton;

typedef struct physis_Texture {
  enum TextureType texture_type;
  uint32_t width;
  uint32_t height;
  uint32_t depth;
  uint32_t rgba_size;
  uint8_t *rgba;
} physis_Texture;

typedef struct ShaderKey {
  uint32_t category;
  uint32_t value;
} ShaderKey;

typedef struct Constant {
  uint32_t id;
  uint32_t num_values;
  float values[4];
} Constant;

typedef struct Sampler {
  TextureUsage texture_usage;
  uint32_t flags;
  uint8_t texture_index;
  uint8_t unknown1;
  uint8_t unknown2;
  uint8_t unknown3;
} Sampler;

typedef struct ColorTableRow {
  float diffuse_color[3];
  float specular_strength;
  float specular_color[3];
  float gloss_strength;
  float emissive_color[3];
  uint16_t tile_set;
  float material_repeat[2];
  float material_skew[2];
} ColorTableRow;

typedef struct physis_ColorTable {
  uint32_t num_rows;
  struct ColorTableRow *rows;
} physis_ColorTable;

typedef struct physis_Material {
  const char *shpk_name;
  uint32_t num_textures;
  const char **textures;
  uint32_t num_shader_keys;
  struct ShaderKey *shader_keys;
  uint32_t num_constants;
  struct Constant *constants;
  uint32_t num_samplers;
  struct Sampler *samplers;
  struct physis_ColorTable color_table;
} physis_Material;

typedef struct physis_ShaderParameter {
  uint16_t slot;
  const char *name;
} physis_ShaderParameter;

typedef struct physis_Shader {
  int32_t len;
  uint8_t *bytecode;
  int32_t num_scalar_parameters;
  struct physis_ShaderParameter *scalar_parameters;
  int32_t num_resource_parameters;
  struct physis_ShaderParameter *resource_parameters;
} physis_Shader;

typedef struct Key {
  uint32_t id;
  uint32_t default_value;
} Key;

typedef struct MaterialParameter {
  uint32_t id;
  uint16_t byte_offset;
  uint16_t byte_size;
} MaterialParameter;

typedef struct physis_SHPK {
  struct ShaderPackage *p_ptr;
  int32_t num_vertex_shaders;
  struct physis_Shader *vertex_shaders;
  int32_t num_pixel_shaders;
  struct physis_Shader *pixel_shaders;
  int32_t num_material_keys;
  struct Key *material_keys;
  uint32_t sub_view_key1_default;
  uint32_t sub_view_key2_default;
  uint32_t material_parameters_size;
  uint32_t num_material_parameters;
  struct MaterialParameter *material_parameters;
} physis_SHPK;

typedef struct Pass {
  uint32_t id;
  uint32_t vertex_shader;
  uint32_t pixel_shader;
} Pass;

typedef struct physis_SHPKNode {
  uint32_t selector;
  uint32_t pass_count;
  uint8_t pass_indices[16];
  uint32_t system_key_count;
  uint32_t *system_keys;
  uint32_t scene_key_count;
  uint32_t *scene_keys;
  uint32_t material_key_count;
  uint32_t *material_keys;
  uint32_t subview_key_count;
  uint32_t *subview_keys;
  struct Pass *passes;
} physis_SHPKNode;

typedef struct physis_CMP {
  struct CMP *p_ptr;
} physis_CMP;

typedef struct RacialScalingParameters {
  /**
   * Male minimum height
   */
  float male_min_size;
  /**
   * Male maximum height
   */
  float male_max_size;
  /**
   * Male minimum tail size
   */
  float male_min_tail;
  /**
   * Male maximum tail size
   */
  float male_max_tail;
  /**
   * Female minimum height
   */
  float female_min_size;
  /**
   * Female maximum height
   */
  float female_max_size;
  /**
   * Female minimum tail size
   */
  float female_min_tail;
  /**
   * Female maximum tail size
   */
  float female_max_tail;
  /**
   * Minimum bust size on the X-axis
   */
  float bust_min_x;
  /**
   * Minimum bust size on the Y-axis
   */
  float bust_min_y;
  /**
   * Minimum bust size on the Z-axis
   */
  float bust_min_z;
  /**
   * Maximum bust size on the X-axis
   */
  float bust_max_x;
  /**
   * Maximum bust size on the Y-axis
   */
  float bust_max_y;
  /**
   * Maximum bust size on the Z-axis
   */
  float bust_max_z;
} RacialScalingParameters;

/**
 * Represents the several options that make up a character data file (DAT) which is used by the game's character creation system to save and load presets.
 */
typedef struct CharacterData {
  /**
   * The version of the character data, the only supported version right now is 4.
   */
  uint32_t version;
  /**
   * The checksum of the data fields.
   */
  uint32_t checksum;
  /**
   * The race of the character.
   */
  Race race;
  /**
   * The gender of the character.
   */
  Gender gender;
  /**
   * The age of the character. Normal = 1, Old = 3, Young = 4.
   */
  uint8_t age;
  /**
   * The height of the character.
   */
  uint8_t height;
  /**
   * The character's subrace.
   */
  Subrace subrace;
  /**
   * The character's selected head.
   */
  uint8_t head;
  /**
   * The character's selected hair.
   */
  uint8_t hair;
  /**
   * If hair highlights are enabled for this character.
   */
  bool enable_highlights;
  /**
   * The character's skin tone.
   */
  uint8_t skin_tone;
  /**
   * The character's right eye color.
   */
  uint8_t right_eye_color;
  /**
   * The character's hair color.
   */
  uint8_t hair_tone;
  /**
   * The color of the hair highlights.
   */
  uint8_t highlights;
  /**
   * The selected facial features.
   */
  uint8_t facial_features;
  /**
   * If the character has limbal eyes.
   */
  uint8_t limbal_eyes;
  /**
   * The character's selected eyebrows.
   */
  uint8_t eyebrows;
  /**
   * The character's left eye color.
   */
  uint8_t left_eye_color;
  /**
   * The character's selected eyes.
   */
  uint8_t eyes;
  /**
   * The character's selected nose.
   */
  uint8_t nose;
  /**
   * The character's selected jaw.
   */
  uint8_t jaw;
  /**
   * The character's selected mouth.
   */
  uint8_t mouth;
  /**
   * The character's selected pattern.
   */
  uint8_t lips_tone_fur_pattern;
  /**
   * The character's selected tail.
   */
  uint8_t tail;
  /**
   * The character's choice of face paint.
   */
  uint8_t face_paint;
  /**
   * The size of the character's bust.
   */
  uint8_t bust;
  /**
   * The color of the face paint.
   */
  uint8_t face_paint_color;
  /**
   * The character's chosen voice.
   */
  uint8_t voice;
  /**
   * The timestamp when the preset was created.
   */
  uint8_t timestamp[4];
} CharacterData;

typedef struct physis_ConfigFile {
  struct ConfigFile *p_ptr;
} physis_ConfigFile;

typedef struct physis_PBD {
  struct PreBoneDeformer *p_ptr;
} physis_PBD;

typedef struct physis_PreBoneDeformBone {
  const char *name;
  float deform[12];
} physis_PreBoneDeformBone;

typedef struct physis_PreBoneDeformMatrices {
  int32_t num_bones;
  struct physis_PreBoneDeformBone *bones;
} physis_PreBoneDeformMatrices;

typedef struct physis_PlateModel {
  float position[2];
  const char *filename;
} physis_PlateModel;

typedef struct physis_Terrain {
  int32_t num_plates;
  struct physis_PlateModel *plates;
} physis_Terrain;

typedef struct physis_Dictionary {
  int32_t num_words;
  const char *const *words;
} physis_Dictionary;

typedef struct physis_IndexEntries {
  uint32_t num_entries;
  const uint32_t *dir_entries;
  const uint32_t *filename_entries;
} physis_IndexEntries;

#if defined(ENABLE_LOGGING)
typedef void (*LogCallback)(QtMsgType, const char*, const char*, int32_t);
#endif

typedef struct physis_SCHD {
  struct Schd *p_ptr;
  enum ShaderStage shader_stage;
} physis_SCHD;

/**
 * An existing install location on disk
 */
typedef struct physis_ExistingGameDirectory {
  /**
   * The application where this installation was from
   */
  enum ExistingInstallType install_type;
  /**
   * The path to the "main folder" where "game" and "boot" sits
   */
  const char *path;
} physis_ExistingGameDirectory;

typedef struct physis_ExistingGameDirectories {
  uint32_t count;
  struct physis_ExistingGameDirectory *entries;
} physis_ExistingGameDirectories;

/**
 * An existing user directory
 */
typedef struct physis_ExistingUserDirectory {
  /**
   * The application where this directory was from
   */
  enum ExistingInstallType install_type;
  /**
   * The path to the user folder
   */
  const char *path;
} physis_ExistingUserDirectory;

typedef struct physis_ExistingUserDirectories {
  uint32_t count;
  struct physis_ExistingUserDirectory *entries;
} physis_ExistingUserDirectories;

const char *physis_get_physis_version(void);

const char *physis_get_libphysis_version(void);

struct physis_Buffer physis_read_file(const char *path);

/**
 * Checks if the file at `path` exists.
 */
bool physis_gamedata_exists(struct GameData *game_data, const char *path);

void physis_gamedata_free_repositories(struct physis_Repositories repositories);

/**
 * Extracts the raw game file from `path`, and puts it in `data` with `size` length. If the path was not found,
 * `size` is 0 and `data` is NULL.
 */
struct physis_Buffer physis_gamedata_extract_file(struct GameData *game_data,
                                                  const char *path);

uint32_t physis_gamedata_find_offset(struct GameData *game_data, const char *path);

void physis_gamedata_free_sheet_header(struct physis_EXH*);

/**
 * Initializes a new GameData structure. Path must be a valid game path, or else it will return NULL.
 */
struct GameData *physis_gamedata_initialize(const char *path);

void physis_gamedata_free(struct GameData *game_data);

struct physis_Repositories physis_gamedata_get_repositories(const struct GameData *game_data);

struct physis_EXD physis_gamedata_read_excel_sheet(struct GameData *game_data,
                                                   const char *name,
                                                   const struct physis_EXH *exh,
                                                   Language language,
                                                   unsigned int page);

const char *physis_gamedata_get_exd_filename(const char *name,
                                             const struct physis_EXH *exh,
                                             Language language,
                                             unsigned int page);

void physis_gamedata_free_sheet(struct physis_EXD exd);

bool physis_gamedata_apply_patch(const struct GameData *gamedata, const char *path);

struct physis_SheetNames physis_gamedata_get_all_sheet_names(struct GameData *game_data);

const char *physis_bootdata_get_version(const struct BootData *boot_data);

/**
 * Initializes a new BootData structure. Path must be a valid boot path, or else it will return NULL.
 */
struct BootData *physis_bootdata_initialize(const char *path);

void physis_bootdata_free(struct BootData *boot_data);

bool physis_bootdata_apply_patch(const struct BootData *bootdata, const char *path);

struct physis_MDL physis_mdl_parse(struct physis_Buffer buffer);

struct physis_Buffer physis_mdl_write(const struct physis_MDL *mdl);

void physis_mdl_replace_vertices(struct physis_MDL *mdl,
                                 uint32_t lod_index,
                                 uint32_t part_index,
                                 uint32_t num_vertices,
                                 const struct Vertex *vertices_ptr,
                                 uint32_t num_indices,
                                 const uint16_t *indices_ptr,
                                 uint32_t num_submeshes,
                                 const struct SubMesh *submeshes_ptr);

void physis_mdl_remove_shape_meshes(struct physis_MDL *mdl);

void physis_mdl_add_shape_mesh(struct physis_MDL *mdl,
                               uint32_t lod_index,
                               uint32_t shape_index,
                               uint32_t shape_mesh_index,
                               uint32_t part_index,
                               uint32_t num_shape_values,
                               const struct NewShapeValue *shape_values);

int32_t physis_get_race_code(Race race, Subrace subrace, Gender gender);

struct physis_SupportedSubraces physis_get_supported_subraces(Race race);

struct physis_EXL physis_gamedata_read_excel_list(struct physis_Buffer buffer);

Slot physis_slot_from_id(int32_t slot_id);

const char *physis_get_slot_name(Slot slot);

const char *physis_build_equipment_path(int32_t model_id,
                                        Race race,
                                        Subrace subrace,
                                        Gender gender,
                                        Slot slot);

const char *physis_build_character_path(CharacterCategory category,
                                        int32_t body_ver,
                                        Race race,
                                        Subrace subrace,
                                        Gender gender);

const char *physis_build_skin_material_path(int32_t race_code,
                                            int32_t body_code,
                                            const char *material_name);

const char *physis_build_gear_material_path(int32_t gear_id,
                                            int32_t gear_version,
                                            const char *material_name);

const char *physis_build_face_material_path(int32_t race_code,
                                            int32_t face_code,
                                            const char *material_name);

const char *physis_build_hair_material_path(int32_t race_code,
                                            int32_t hair_code,
                                            const char *material_name);

const char *physis_build_ear_material_path(int32_t race_code,
                                           int32_t ear_code,
                                           const char *material_name);

const char *physis_build_tail_material_path(int32_t race_code,
                                            int32_t tail_code,
                                            const char *material_name);

struct Blowfish *physis_blowfish_initialize(uint8_t *key, unsigned int key_size);

void physis_blowfish_free(struct Blowfish *blowfish);

bool physis_blowfish_encrypt(const struct Blowfish *blowfish,
                             uint8_t *in_data,
                             unsigned int in_data_size,
                             uint8_t **out_data,
                             uint32_t *out_data_size);

bool physis_blowfish_decrypt(const struct Blowfish *blowfish,
                             uint8_t *in_data,
                             unsigned int in_data_size,
                             uint8_t **out_data,
                             uint32_t *out_data_size);

bool physis_install_game(const char *installer_path, const char *game_directory);

struct physis_EXH *physis_parse_excel_sheet_header(struct physis_Buffer buffer);

struct physis_Skeleton physis_parse_skeleton(struct physis_Buffer buffer);

struct physis_Texture physis_texture_parse(struct physis_Buffer buffer);

struct physis_Material physis_material_parse(struct physis_Buffer buffer);

struct physis_SHPK physis_parse_shpk(struct physis_Buffer buffer);

struct physis_SHPKNode physis_shpk_get_node(const struct physis_SHPK *shpk, uint32_t key);

uint32_t physis_shpk_build_selector_from_all_keys(const uint32_t *system_keys,
                                                  uint32_t system_key_count,
                                                  const uint32_t *scene_keys,
                                                  uint32_t scene_key_count,
                                                  const uint32_t *material_keys,
                                                  uint32_t material_key_count,
                                                  const uint32_t *subview_keys,
                                                  uint32_t subview_key_count);

uint32_t physis_shpk_crc(const char *name);

struct physis_CMP physis_cmp_parse(struct physis_Buffer buffer);

struct RacialScalingParameters physis_cmp_get_racial_scaling_parameters(struct physis_CMP cmp,
                                                                        Race,
                                                                        Subrace subrace);

struct CharacterData physis_chardat_parse(struct physis_Buffer buffer);

struct physis_ConfigFile physis_cfg_parse(struct physis_Buffer buffer);

void physis_cfg_set_value(struct physis_ConfigFile cfg, const char *key, const char *value);

struct physis_Buffer physis_cfg_write(struct physis_ConfigFile cfg);

struct physis_PBD physis_parse_pbd(struct physis_Buffer buffer);

struct physis_PreBoneDeformMatrices physis_pbd_get_deform_matrix(struct physis_PBD pbd,
                                                                 uint16_t from_body_id,
                                                                 uint16_t to_body_id);

struct physis_Terrain physis_parse_tera(struct physis_Buffer buffer);

struct physis_Dictionary physis_parse_dictionary(struct physis_Buffer buffer);

struct physis_IndexEntries physis_index_parse(const char *path);

uint32_t physis_generate_partial_hash(const char *name);

uint64_t physis_calculate_hash(const char *path);

#if defined(ENABLE_LOGGING)
void set_tracing_callback(LogCallback callback);
#endif

#if defined(ENABLE_LOGGING)
void physis_initialize_logging(void);
#endif

struct physis_SCHD physis_parse_schd(struct physis_Buffer _buffer);

struct physis_ExistingGameDirectories physis_find_existing_game_dirs(void);

struct physis_ExistingUserDirectories physis_find_existing_user_dirs(void);
