#pragma once

#include <cstdarg>
#include <cstdint>
#include <cstdlib>

enum class CharacterCategory : uint8_t {
  Body,
  Hair,
  Face,
  Tail,
  Ear,
};

/// Where the existing installation came from
enum class ExistingInstallType {
  /// Installed via the official launcher
  OfficialLauncher,
  /// Installed via XIVQuickLauncher
  XIVQuickLauncher,
  /// Installed via XIVLauncherCore
  XIVLauncherCore,
  /// Installed via XIVOnMac
  XIVOnMac,
  /// Installed via Astra
  Astra,
};

/// Gender of the character.
enum class Gender : uint8_t {
  Male,
  Female,
};

/// The language the game data is written for. Some of these languages are supported in the Global region.
enum class Language : uint8_t {
  /// Used for data that is language-agnostic, such as item data.
  None,
  /// Japanese language.
  Japanese,
  /// English language.
  English,
  /// German language.
  German,
  /// French language.
  French,
  /// Chinese (Simplified) language.
  ChineseSimplified,
  /// Chinese (Traditional) language.
  ChineseTraditional,
  /// Korean language.
  Korean,
};

/// The major races of Eorzea.
enum class Race : uint8_t {
  Hyur,
  Elezen,
  Lalafell,
  Miqote,
  Roegadyn,
  AuRa,
  Hrothgar,
  Viera,
};

enum class ShaderStage {
  Vertex,
  Pixel,
};

/// The slot the item is for.
enum class Slot : uint8_t {
  /// The head slot. Shorthand is "met".
  Head,
  /// The hands slot. Shorthand is "glv".
  Hands,
  /// The legs slot. Shorthand is "dwn".
  Legs,
  /// The feet slot. Shorthand is "sho".
  Feet,
  /// The body or chest slot. Shorthand is "top".
  Body,
  /// The earrings slot. Shorthand is "ear".
  Earring,
  /// The neck slot. Shorthand is "nek".
  Neck,
  /// The ring slot. Shorthand is "rir".
  Rings,
  /// The wrists slot. Shorthand is "wrs".
  Wrists,
};

/// The race's "subrace". Each race has two subraces, which are actually identical (even down to the ids!)
/// with the exception of Hyurs, which have two unique subraces that are really two separate races.
enum class Subrace : uint8_t {
  Midlander,
  Highlander,
  Wildwood,
  Duskwight,
  Plainsfolk,
  Dunesfolk,
  Seeker,
  Keeper,
  SeaWolf,
  Hellsguard,
  Raen,
  Xaela,
  Hellion,
  Lost,
  Rava,
  Veena,
};

enum class TextureType {
  TwoDimensional,
  ThreeDimensional,
};

enum class TextureUsage : uint8_t {
  Sampler,
  Sampler0,
  Sampler1,
  SamplerCatchlight,
  SamplerColorMap0,
  SamplerColorMap1,
  SamplerDiffuse,
  SamplerEnvMap,
  SamplerMask,
  SamplerNormal,
  SamplerNormalMap0,
  SamplerNormalMap1,
  SamplerReflection,
  SamplerSpecular,
  SamplerSpecularMap0,
  SamplerSpecularMap1,
  SamplerWaveMap,
  SamplerWaveletMap0,
  SamplerWaveletMap1,
  SamplerWhitecapMap,
  UnknownDawntrail1,
};

/// Implementation of the Blowfish block cipher, specialized for encrypting and decrypting SqexArg - the technique used to encrypt game arguments by the launcher.
///
/// # Example
///
/// ```
/// # use physis::blowfish::Blowfish;
/// let key = b"abcdefgh";
/// let data = b"foobar  ";
///
/// let fish = Blowfish::new(key);
/// let encrypted = fish.encrypt(data).unwrap();
/// let decrypted = fish.decrypt(&encrypted).unwrap();
/// # assert_eq!(data, &decrypted[..])
/// ```
struct Blowfish;

/// Represents the boot data for FFXIV, which is located under the "boot" directory.
struct BootData;

struct CMP;

/// Represents a config file, which is made up of categories and settings. Categories may have zero to one setting.
struct ConfigFile;

struct EXD;

struct EXH;

/// Framework for operating on game data.
struct GameData;

struct MDL;

struct PreBoneDeformer;

struct Schd;

struct ShaderPackage;

struct physis_Buffer {
  uint32_t size;
  uint8_t *data;
};

/// The type of repository, discerning game data from expansion data.
struct RepositoryType {
  enum class Tag {
    /// The base game directory, like "ffxiv".
    Base,
    /// An expansion directory, like "ex1".
    Expansion,
  };

  struct Expansion_Body {
    /// The expansion number starting at 1.
    int32_t number;
  };

  Tag tag;
  union {
    Expansion_Body expansion;
  };
};

struct physis_Repository {
  const char *name;
  RepositoryType repository_type;
  const char *version;
};

struct physis_Repositories {
  uint32_t repositories_count;
  physis_Repository *repositories;
};

struct physis_EXH {
  EXH *p_ptr;
  uint32_t page_count;
  uint32_t language_count;
  Language *languages;
  uint32_t column_count;
  uint32_t row_count;
};

struct physis_ColumnData {
  enum class Tag {
    String,
    Bool,
    Int8,
    UInt8,
    Int16,
    UInt16,
    Int32,
    UInt32,
    Float32,
    Int64,
    UInt64,
  };

  struct String_Body {
    const char *_0;
  };

  struct Bool_Body {
    bool _0;
  };

  struct Int8_Body {
    int8_t _0;
  };

  struct UInt8_Body {
    uint8_t _0;
  };

  struct Int16_Body {
    int16_t _0;
  };

  struct UInt16_Body {
    uint16_t _0;
  };

  struct Int32_Body {
    int32_t _0;
  };

  struct UInt32_Body {
    uint32_t _0;
  };

  struct Float32_Body {
    float _0;
  };

  struct Int64_Body {
    int64_t _0;
  };

  struct UInt64_Body {
    uint64_t _0;
  };

  Tag tag;
  union {
    String_Body string;
    Bool_Body bool_;
    Int8_Body int8;
    UInt8_Body u_int8;
    Int16_Body int16;
    UInt16_Body u_int16;
    Int32_Body int32;
    UInt32_Body u_int32;
    Float32_Body float32;
    Int64_Body int64;
    UInt64_Body u_int64;
  };
};

struct physis_ExcelRow {
  physis_ColumnData *column_data;
};

struct physis_EXD {
  EXD *p_ptr;
  unsigned int column_count;
  physis_ExcelRow *row_data;
  unsigned int row_count;
};

struct physis_SheetNames {
  uint32_t name_count;
  const char **names;
};

struct Vertex {
  float position[3];
  float uv0[2];
  float uv1[2];
  float normal[3];
  float bitangent[4];
  float color[4];
  float bone_weight[4];
  uint8_t bone_id[4];
};

struct SubMesh {
  uintptr_t submesh_index;
  uint32_t index_count;
  uint32_t index_offset;
};

struct physis_Shape {
  const char *name;
  const Vertex *morphed_vertices;
};

struct physis_Part {
  uint32_t num_vertices;
  const Vertex *vertices;
  uint32_t num_indices;
  const uint16_t *indices;
  uint16_t material_index;
  uint32_t num_submeshes;
  const SubMesh *submeshes;
  uint32_t num_shapes;
  physis_Shape *shapes;
};

struct physis_LOD {
  uint32_t num_parts;
  const physis_Part *parts;
};

struct physis_MDL {
  MDL *p_ptr;
  uint32_t num_lod;
  const physis_LOD *lods;
  uint32_t num_affected_bones;
  const char **affected_bone_names;
  uint32_t num_material_names;
  const char **material_names;
};

struct NewShapeValue {
  uint32_t base_index;
  Vertex replacing_vertex;
};

struct physis_SupportedSubraces {
  Subrace subraces[2];
};

struct physis_EXL {
  int32_t version;
  int32_t entry_count;
  const char *const *entry_keys;
  const int32_t *entry_values;
};

struct physis_Bone {
  uint32_t index;
  const char *name;
  physis_Bone *parent_bone;
  uint32_t parent_index;
  float position[3];
  float rotation[4];
  float scale[3];
};

struct physis_Skeleton {
  uint32_t num_bones;
  physis_Bone *bones;
  physis_Bone *root_bone;
};

struct physis_Texture {
  TextureType texture_type;
  uint32_t width;
  uint32_t height;
  uint32_t depth;
  uint32_t rgba_size;
  uint8_t *rgba;
};

struct ShaderKey {
  uint32_t category;
  uint32_t value;
};

struct Constant {
  uint32_t id;
  uint32_t num_values;
  float values[4];
};

struct Sampler {
  TextureUsage texture_usage;
  uint32_t flags;
  uint8_t texture_index;
  uint8_t unknown1;
  uint8_t unknown2;
  uint8_t unknown3;
};

struct ColorTableRow {
  float diffuse_color[3];
  float specular_strength;
  float specular_color[3];
  float gloss_strength;
  float emissive_color[3];
  uint16_t tile_set;
  float material_repeat[2];
  float material_skew[2];
};

struct physis_ColorTable {
  uint32_t num_rows;
  ColorTableRow *rows;
};

struct physis_Material {
  const char *shpk_name;
  uint32_t num_textures;
  const char **textures;
  uint32_t num_shader_keys;
  ShaderKey *shader_keys;
  uint32_t num_constants;
  Constant *constants;
  uint32_t num_samplers;
  Sampler *samplers;
  physis_ColorTable color_table;
};

struct physis_ShaderParameter {
  uint16_t slot;
  const char *name;
};

struct physis_Shader {
  int32_t len;
  uint8_t *bytecode;
  int32_t num_scalar_parameters;
  physis_ShaderParameter *scalar_parameters;
  int32_t num_resource_parameters;
  physis_ShaderParameter *resource_parameters;
};

struct Key {
  uint32_t id;
  uint32_t default_value;
};

struct MaterialParameter {
  uint32_t id;
  uint16_t byte_offset;
  uint16_t byte_size;
};

struct physis_SHPK {
  ShaderPackage *p_ptr;
  int32_t num_vertex_shaders;
  physis_Shader *vertex_shaders;
  int32_t num_pixel_shaders;
  physis_Shader *pixel_shaders;
  int32_t num_material_keys;
  Key *material_keys;
  uint32_t sub_view_key1_default;
  uint32_t sub_view_key2_default;
  uint32_t material_parameters_size;
  uint32_t num_material_parameters;
  MaterialParameter *material_parameters;
};

struct Pass {
  uint32_t id;
  uint32_t vertex_shader;
  uint32_t pixel_shader;
};

struct physis_SHPKNode {
  uint32_t selector;
  uint32_t pass_count;
  uint8_t pass_indices[16];
  uint32_t system_key_count;
  uint32_t *system_keys;
  uint32_t scene_key_count;
  uint32_t *scene_keys;
  uint32_t material_key_count;
  uint32_t *material_keys;
  uint32_t subview_key_count;
  uint32_t *subview_keys;
  Pass *passes;
};

struct physis_CMP {
  CMP *p_ptr;
};

struct RacialScalingParameters {
  /// Male minimum height
  float male_min_size;
  /// Male maximum height
  float male_max_size;
  /// Male minimum tail size
  float male_min_tail;
  /// Male maximum tail size
  float male_max_tail;
  /// Female minimum height
  float female_min_size;
  /// Female maximum height
  float female_max_size;
  /// Female minimum tail size
  float female_min_tail;
  /// Female maximum tail size
  float female_max_tail;
  /// Minimum bust size on the X-axis
  float bust_min_x;
  /// Minimum bust size on the Y-axis
  float bust_min_y;
  /// Minimum bust size on the Z-axis
  float bust_min_z;
  /// Maximum bust size on the X-axis
  float bust_max_x;
  /// Maximum bust size on the Y-axis
  float bust_max_y;
  /// Maximum bust size on the Z-axis
  float bust_max_z;
};

/// Represents the several options that make up a character data file (DAT) which is used by the game's character creation system to save and load presets.
struct CharacterData {
  /// The version of the character data, the only supported version right now is 4.
  uint32_t version;
  /// The checksum of the data fields.
  uint32_t checksum;
  /// The race of the character.
  Race race;
  /// The gender of the character.
  Gender gender;
  /// The age of the character. Normal = 1, Old = 3, Young = 4.
  uint8_t age;
  /// The height of the character.
  uint8_t height;
  /// The character's subrace.
  Subrace subrace;
  /// The character's selected head.
  uint8_t head;
  /// The character's selected hair.
  uint8_t hair;
  /// If hair highlights are enabled for this character.
  bool enable_highlights;
  /// The character's skin tone.
  uint8_t skin_tone;
  /// The character's right eye color.
  uint8_t right_eye_color;
  /// The character's hair color.
  uint8_t hair_tone;
  /// The color of the hair highlights.
  uint8_t highlights;
  /// The selected facial features.
  uint8_t facial_features;
  /// If the character has limbal eyes.
  uint8_t limbal_eyes;
  /// The character's selected eyebrows.
  uint8_t eyebrows;
  /// The character's left eye color.
  uint8_t left_eye_color;
  /// The character's selected eyes.
  uint8_t eyes;
  /// The character's selected nose.
  uint8_t nose;
  /// The character's selected jaw.
  uint8_t jaw;
  /// The character's selected mouth.
  uint8_t mouth;
  /// The character's selected pattern.
  uint8_t lips_tone_fur_pattern;
  /// The character's selected tail.
  uint8_t tail;
  /// The character's choice of face paint.
  uint8_t face_paint;
  /// The size of the character's bust.
  uint8_t bust;
  /// The color of the face paint.
  uint8_t face_paint_color;
  /// The character's chosen voice.
  uint8_t voice;
  /// The timestamp when the preset was created.
  uint8_t timestamp[4];
};

struct physis_ConfigFile {
  ConfigFile *p_ptr;
};

struct physis_PBD {
  PreBoneDeformer *p_ptr;
};

struct physis_PreBoneDeformBone {
  const char *name;
  float deform[12];
};

struct physis_PreBoneDeformMatrices {
  int32_t num_bones;
  physis_PreBoneDeformBone *bones;
};

struct physis_PlateModel {
  float position[2];
  const char *filename;
};

struct physis_Terrain {
  int32_t num_plates;
  physis_PlateModel *plates;
};

struct physis_Dictionary {
  int32_t num_words;
  const char *const *words;
};

struct physis_IndexEntries {
  uint32_t num_entries;
  const uint32_t *dir_entries;
  const uint32_t *filename_entries;
};

#if defined(ENABLE_LOGGING)
using LogCallback = void(*)(QtMsgType, const char*, const char*, int32_t);
#endif

struct physis_SCHD {
  Schd *p_ptr;
  ShaderStage shader_stage;
};

/// An existing install location on disk
struct physis_ExistingGameDirectory {
  /// The application where this installation was from
  ExistingInstallType install_type;
  /// The path to the "main folder" where "game" and "boot" sits
  const char *path;
};

struct physis_ExistingGameDirectories {
  uint32_t count;
  physis_ExistingGameDirectory *entries;
};

/// An existing user directory
struct physis_ExistingUserDirectory {
  /// The application where this directory was from
  ExistingInstallType install_type;
  /// The path to the user folder
  const char *path;
};

struct physis_ExistingUserDirectories {
  uint32_t count;
  physis_ExistingUserDirectory *entries;
};

extern "C" {

const char *physis_get_physis_version();

const char *physis_get_libphysis_version();

physis_Buffer physis_read_file(const char *path);

/// Checks if the file at `path` exists.
bool physis_gamedata_exists(GameData *game_data, const char *path);

void physis_gamedata_free_repositories(physis_Repositories repositories);

/// Extracts the raw game file from `path`, and puts it in `data` with `size` length. If the path was not found,
/// `size` is 0 and `data` is NULL.
physis_Buffer physis_gamedata_extract_file(GameData *game_data,
                                           const char *path);

uint32_t physis_gamedata_find_offset(GameData *game_data, const char *path);

void physis_gamedata_free_sheet_header(physis_EXH*);

/// Initializes a new GameData structure. Path must be a valid game path, or else it will return NULL.
GameData *physis_gamedata_initialize(const char *path);

void physis_gamedata_free(GameData *game_data);

physis_Repositories physis_gamedata_get_repositories(const GameData *game_data);

physis_EXD physis_gamedata_read_excel_sheet(GameData *game_data,
                                            const char *name,
                                            const physis_EXH *exh,
                                            Language language,
                                            unsigned int page);

const char *physis_gamedata_get_exd_filename(const char *name,
                                             const physis_EXH *exh,
                                             Language language,
                                             unsigned int page);

void physis_gamedata_free_sheet(physis_EXD exd);

bool physis_gamedata_apply_patch(const GameData *gamedata, const char *path);

physis_SheetNames physis_gamedata_get_all_sheet_names(GameData *game_data);

const char *physis_bootdata_get_version(const BootData *boot_data);

/// Initializes a new BootData structure. Path must be a valid boot path, or else it will return NULL.
BootData *physis_bootdata_initialize(const char *path);

void physis_bootdata_free(BootData *boot_data);

bool physis_bootdata_apply_patch(const BootData *bootdata, const char *path);

physis_MDL physis_mdl_parse(physis_Buffer buffer);

physis_Buffer physis_mdl_write(const physis_MDL *mdl);

void physis_mdl_replace_vertices(physis_MDL *mdl,
                                 uint32_t lod_index,
                                 uint32_t part_index,
                                 uint32_t num_vertices,
                                 const Vertex *vertices_ptr,
                                 uint32_t num_indices,
                                 const uint16_t *indices_ptr,
                                 uint32_t num_submeshes,
                                 const SubMesh *submeshes_ptr);

void physis_mdl_remove_shape_meshes(physis_MDL *mdl);

void physis_mdl_add_shape_mesh(physis_MDL *mdl,
                               uint32_t lod_index,
                               uint32_t shape_index,
                               uint32_t shape_mesh_index,
                               uint32_t part_index,
                               uint32_t num_shape_values,
                               const NewShapeValue *shape_values);

int32_t physis_get_race_code(Race race, Subrace subrace, Gender gender);

physis_SupportedSubraces physis_get_supported_subraces(Race race);

physis_EXL physis_gamedata_read_excel_list(physis_Buffer buffer);

Slot physis_slot_from_id(int32_t slot_id);

const char *physis_get_slot_name(Slot slot);

const char *physis_build_equipment_path(int32_t model_id,
                                        Race race,
                                        Subrace subrace,
                                        Gender gender,
                                        Slot slot);

const char *physis_build_character_path(CharacterCategory category,
                                        int32_t body_ver,
                                        Race race,
                                        Subrace subrace,
                                        Gender gender);

const char *physis_build_skin_material_path(int32_t race_code,
                                            int32_t body_code,
                                            const char *material_name);

const char *physis_build_gear_material_path(int32_t gear_id,
                                            int32_t gear_version,
                                            const char *material_name);

const char *physis_build_face_material_path(int32_t race_code,
                                            int32_t face_code,
                                            const char *material_name);

const char *physis_build_hair_material_path(int32_t race_code,
                                            int32_t hair_code,
                                            const char *material_name);

const char *physis_build_ear_material_path(int32_t race_code,
                                           int32_t ear_code,
                                           const char *material_name);

const char *physis_build_tail_material_path(int32_t race_code,
                                            int32_t tail_code,
                                            const char *material_name);

Blowfish *physis_blowfish_initialize(uint8_t *key, unsigned int key_size);

void physis_blowfish_free(Blowfish *blowfish);

bool physis_blowfish_encrypt(const Blowfish *blowfish,
                             uint8_t *in_data,
                             unsigned int in_data_size,
                             uint8_t **out_data,
                             uint32_t *out_data_size);

bool physis_blowfish_decrypt(const Blowfish *blowfish,
                             uint8_t *in_data,
                             unsigned int in_data_size,
                             uint8_t **out_data,
                             uint32_t *out_data_size);

bool physis_install_game(const char *installer_path, const char *game_directory);

physis_EXH *physis_parse_excel_sheet_header(physis_Buffer buffer);

physis_Skeleton physis_parse_skeleton(physis_Buffer buffer);

physis_Texture physis_texture_parse(physis_Buffer buffer);

physis_Material physis_material_parse(physis_Buffer buffer);

physis_SHPK physis_parse_shpk(physis_Buffer buffer);

physis_SHPKNode physis_shpk_get_node(const physis_SHPK *shpk, uint32_t key);

uint32_t physis_shpk_build_selector_from_all_keys(const uint32_t *system_keys,
                                                  uint32_t system_key_count,
                                                  const uint32_t *scene_keys,
                                                  uint32_t scene_key_count,
                                                  const uint32_t *material_keys,
                                                  uint32_t material_key_count,
                                                  const uint32_t *subview_keys,
                                                  uint32_t subview_key_count);

uint32_t physis_shpk_crc(const char *name);

physis_CMP physis_cmp_parse(physis_Buffer buffer);

RacialScalingParameters physis_cmp_get_racial_scaling_parameters(physis_CMP cmp,
                                                                 Race,
                                                                 Subrace subrace);

CharacterData physis_chardat_parse(physis_Buffer buffer);

physis_ConfigFile physis_cfg_parse(physis_Buffer buffer);

void physis_cfg_set_value(physis_ConfigFile cfg, const char *key, const char *value);

physis_Buffer physis_cfg_write(physis_ConfigFile cfg);

physis_PBD physis_parse_pbd(physis_Buffer buffer);

physis_PreBoneDeformMatrices physis_pbd_get_deform_matrix(physis_PBD pbd,
                                                          uint16_t from_body_id,
                                                          uint16_t to_body_id);

physis_Terrain physis_parse_tera(physis_Buffer buffer);

physis_Dictionary physis_parse_dictionary(physis_Buffer buffer);

physis_IndexEntries physis_index_parse(const char *path);

uint32_t physis_generate_partial_hash(const char *name);

uint64_t physis_calculate_hash(const char *path);

#if defined(ENABLE_LOGGING)
void set_tracing_callback(LogCallback callback);
#endif

#if defined(ENABLE_LOGGING)
void physis_initialize_logging();
#endif

physis_SCHD physis_parse_schd(physis_Buffer _buffer);

physis_ExistingGameDirectories physis_find_existing_game_dirs();

physis_ExistingUserDirectories physis_find_existing_user_dirs();

} // extern "C"
